from math import gcd

import pandas as pd

from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils.data.tabularization import get_shared_times
from darts.utils.timeseries_generation import linear_timeseries


# math.lcm is not available in Python <= 3.8, so we define it here
def lcm(*integers):
    a = integers[0]
    for b in integers[1:]:
        a = (a * b) // gcd(a, b)
    return a


class GetSharedTimesTestCase(DartsBaseTestClass):

    """
    Tests `get_shared_times` function defined in `darts.utils.data.tabularization`.
    """

    def test_shared_times_equal_freq_range_idx(self):
        """
        Tests that `get_shared_times` correctly handles range time
        index series that are of equal frequency.
        """
        # `series_1` begins before `series_2` does and ends
        # before `series_2` does, and `series_2` begins before
        # `series_3` does and ends before `series_3` does:
        series_1 = linear_timeseries(start=1, end=11, freq=2)
        series_2 = linear_timeseries(start=3, end=13, freq=2)
        series_3 = linear_timeseries(start=5, end=15, freq=2)

        # Intersection of a single time index is just the original time index:
        self.assertTrue(series_1.time_index.equals(get_shared_times(series_1)))
        self.assertTrue(series_2.time_index.equals(get_shared_times(series_2)))
        self.assertTrue(series_3.time_index.equals(get_shared_times(series_3)))

        # Intersection of two time indices begins at start time of later series
        # and stops at end time of earlier series.
        # Since `series_1` is before `series_2`:
        expected_12 = linear_timeseries(
            start=series_2.start_time(), end=series_1.end_time(), freq=series_1.freq
        )
        self.assertTrue(
            expected_12.time_index.equals(get_shared_times(series_1, series_2))
        )
        # Since `series_2` is before `series_3`:
        expected_23 = linear_timeseries(
            start=series_3.start_time(), end=series_2.end_time(), freq=series_2.freq
        )
        self.assertTrue(
            expected_23.time_index.equals(get_shared_times(series_2, series_3))
        )
        # Since `series_1` is before `series_3`:
        expected_13 = linear_timeseries(
            start=series_3.start_time(), end=series_1.end_time(), freq=series_1.freq
        )
        self.assertTrue(
            expected_13.time_index.equals(get_shared_times(series_1, series_3))
        )

        # Intersection of all three time series should begin at start of series_3 (i.e.
        # the last series to begin) and end at the end of series_1 (i.e. the first series
        # to end):
        expected_123 = linear_timeseries(
            start=series_3.start_time(), end=series_1.end_time(), freq=series_1.freq
        )
        self.assertTrue(
            expected_123.time_index.equals(
                get_shared_times(series_1, series_2, series_3)
            )
        )

    def test_shared_times_equal_freq_datetime_idx(self):
        """
        Tests that `get_shared_times` correctly handles datetime time
        index series that are of equal frequency.
        """
        # `series_1` begins before `series_2` does and ends
        # before `series_2` does, and `series_2` begins before
        # `series_3` does and ends before `series_3` does:
        series_1 = linear_timeseries(
            start=pd.Timestamp("1/1/2000"), end=pd.Timestamp("1/11/2000"), freq="2d"
        )
        series_2 = linear_timeseries(
            start=pd.Timestamp("1/3/2000"), end=pd.Timestamp("1/13/2000"), freq="2d"
        )
        series_3 = linear_timeseries(
            start=pd.Timestamp("1/5/2000"), end=pd.Timestamp("1/15/2000"), freq="2d"
        )

        # Intersection of a single time index is just the original time index:
        self.assertTrue(series_1.time_index.equals(get_shared_times(series_1)))
        self.assertTrue(series_2.time_index.equals(get_shared_times(series_2)))
        self.assertTrue(series_3.time_index.equals(get_shared_times(series_3)))

        # Intersection of two time indices begins at start time of later series
        # and stops at end time of earlier series.
        # Since `series_1` is before `series_2`:
        expected_12 = linear_timeseries(
            start=series_2.start_time(), end=series_1.end_time(), freq=series_1.freq
        )
        self.assertTrue(
            expected_12.time_index.equals(get_shared_times(series_1, series_2))
        )
        # Since `series_2` is before `series_3`:
        expected_23 = linear_timeseries(
            start=series_3.start_time(), end=series_2.end_time(), freq=series_2.freq
        )
        self.assertTrue(
            expected_23.time_index.equals(get_shared_times(series_2, series_3))
        )
        # Since `series_1` is before `series_3`:
        expected_13 = linear_timeseries(
            start=series_3.start_time(), end=series_1.end_time(), freq=series_1.freq
        )
        self.assertTrue(
            expected_13.time_index.equals(get_shared_times(series_1, series_3))
        )

        # Intersection of all three time series should begin at start of series_3 (i.e.
        # the last series to begin) and end at the end of series_1 (i.e. the first series
        # to end):
        expected_123 = linear_timeseries(
            start=series_3.start_time(), end=series_1.end_time(), freq=series_1.freq
        )
        self.assertTrue(
            expected_123.time_index.equals(
                get_shared_times(series_1, series_2, series_3)
            )
        )

    def test_shared_times_unequal_freq_range_idx(self):
        """
        Tests that `get_shared_times` correctly handles range time
        index series that are of different frequencies.
        """
        # `series_1` begins before `series_2` does and ends
        # before `series_2` does, and `series_2` begins before
        # `series_3` does and ends before `series_3` does. Each
        # series is of a different frequency:
        series_1 = linear_timeseries(start=1, end=11, freq=1)
        series_2 = linear_timeseries(start=3, end=13, freq=2)
        series_3 = linear_timeseries(start=5, end=17, freq=3)

        # Intersection of a single time index is just the original time index:
        self.assertTrue(series_1.time_index.equals(get_shared_times(series_1)))
        self.assertTrue(series_2.time_index.equals(get_shared_times(series_2)))
        self.assertTrue(series_3.time_index.equals(get_shared_times(series_3)))

        # Intersection of two time indices begins at start time of later series
        # and stops at end time of earlier series. The frequency of the intersection
        # is the lowest common multiple between the frequencies of the two series:

        # `series_1` is before `series_2`:
        expected_12 = linear_timeseries(
            start=series_2.start_time(),
            end=series_1.end_time(),
            freq=lcm(series_1.freq, series_2.freq),
        )
        # `linear_timeseries` may have added point beyond specified `end`;
        # remove this point if present:
        if expected_12.time_index[-1] > series_1.end_time():
            expected_12 = expected_12.drop_after(expected_12.time_index[-1])
        self.assertTrue(
            expected_12.time_index.equals(get_shared_times(series_1, series_2))
        )
        # `series_2` is before `series_3`:
        expected_23 = linear_timeseries(
            start=series_3.start_time(),
            end=series_2.end_time(),
            freq=lcm(series_2.freq, series_3.freq),
        )
        # `linear_timeseries` may have added point beyond specified `end`;
        # remove this point if present:
        if expected_23.time_index[-1] > series_2.end_time():
            expected_23 = expected_23.drop_after(expected_23.time_index[-1])
        self.assertTrue(
            expected_23.time_index.equals(get_shared_times(series_2, series_3))
        )
        # `series_1` is before `series_3`:
        expected_13 = linear_timeseries(
            start=series_3.start_time(),
            end=series_1.end_time(),
            freq=lcm(series_1.freq, series_3.freq),
        )
        # `linear_timeseries` may have added point beyond specified `end`;
        # remove this point if present:
        if expected_13.time_index[-1] > series_1.end_time():
            expected_13 = expected_13.drop_after(expected_13.time_index[-1])
        self.assertTrue(
            expected_13.time_index.equals(get_shared_times(series_1, series_3))
        )

        # Intersection of all three time series should begin at start of series_3 (i.e.
        # the last series to begin) and end at the end of series_1 (i.e. the first series
        # to end). The frequency of the intersection should be the lowest common multiple
        # shared by all three frequencies:
        expected_123 = linear_timeseries(
            start=series_3.start_time(),
            end=series_1.end_time(),
            freq=lcm(series_1.freq, series_2.freq, series_3.freq),
        )
        if expected_123.time_index[-1] > series_1.end_time():
            expected_123 = expected_123.drop_after(expected_123.time_index[-1])
        self.assertTrue(
            expected_123.time_index.equals(
                get_shared_times(series_1, series_2, series_3)
            )
        )

    def test_shared_times_unequal_freq_datetime_idx(self):
        """
        Tests that `get_shared_times` correctly handles range time
        index series that are of different frequencies.
        """
        # `series_1` begins before `series_2` does and ends
        # before `series_2` does, and `series_2` begins before
        # `series_3` does and ends before `series_3` does. Each
        # series is of a different frequency:
        series_1 = linear_timeseries(
            start=pd.Timestamp("1/1/2000"), end=pd.Timestamp("1/11/2000"), freq="2d"
        )
        series_2 = linear_timeseries(
            start=pd.Timestamp("1/3/2000"), end=pd.Timestamp("1/13/2000"), freq="2d"
        )
        series_3 = linear_timeseries(
            start=pd.Timestamp("1/5/2000"), end=pd.Timestamp("1/15/2000"), freq="2d"
        )

        # Intersection of a single time index is just the original time index:
        self.assertTrue(series_1.time_index.equals(get_shared_times(series_1)))
        self.assertTrue(series_2.time_index.equals(get_shared_times(series_2)))
        self.assertTrue(series_3.time_index.equals(get_shared_times(series_3)))

        # Intersection of two time indices begins at start time of later series
        # and stops at end time of earlier series. The frequency of the intersection
        # is the lowest common multiple between the frequencies of the two series:

        # `series_1` is before `series_2`:
        freq_12 = f"{lcm(series_1.freq.n, series_2.freq.n)}d"
        expected_12 = linear_timeseries(
            start=series_2.start_time(),
            end=series_1.end_time(),
            freq=freq_12,
        )
        # `linear_timeseries` may have added point beyond specified `end`;
        # remove this point if present:
        if expected_12.time_index[-1] > series_1.end_time():
            expected_12 = expected_12.drop_after(expected_12.time_index[-1])
        self.assertTrue(
            expected_12.time_index.equals(get_shared_times(series_1, series_2))
        )
        # `series_2` is before `series_3`:
        freq_23 = f"{lcm(series_2.freq.n, series_3.freq.n)}d"
        expected_23 = linear_timeseries(
            start=series_3.start_time(),
            end=series_2.end_time(),
            freq=freq_23,
        )
        # `linear_timeseries` may have added point beyond specified `end`;
        # remove this point if present:
        if expected_23.time_index[-1] > series_2.end_time():
            expected_23 = expected_23.drop_after(expected_23.time_index[-1])
        self.assertTrue(
            expected_23.time_index.equals(get_shared_times(series_2, series_3))
        )
        # `series_1` is before `series_3`:
        freq_13 = f"{lcm(series_1.freq.n, series_3.freq.n)}d"
        expected_13 = linear_timeseries(
            start=series_3.start_time(),
            end=series_1.end_time(),
            freq=freq_13,
        )
        # `linear_timeseries` may have added point beyond specified `end`;
        # remove this point if present:
        if expected_13.time_index[-1] > series_1.end_time():
            expected_13 = expected_13.drop_after(expected_13.time_index[-1])
        self.assertTrue(
            expected_13.time_index.equals(get_shared_times(series_1, series_3))
        )

        # Intersection of all three time series should begin at start of series_3 (i.e.
        # the last series to begin) and end at the end of series_1 (i.e. the first series
        # to end). The frequency of the intersection should be the lowest common multiple
        # shared by all three frequencies:
        freq_123 = f"{lcm(series_1.freq.n, series_2.freq.n, series_3.freq.n)}d"
        expected_123 = linear_timeseries(
            start=series_3.start_time(),
            end=series_1.end_time(),
            freq=freq_123,
        )
        if expected_123.time_index[-1] > series_1.end_time():
            expected_123 = expected_123.drop_after(expected_123.time_index[-1])
        self.assertTrue(
            expected_123.time_index.equals(
                get_shared_times(series_1, series_2, series_3)
            )
        )

    def test_shared_times_no_overlap_range_idx(self):
        """
        Tests that `get_shared_times` returns `None` when
        supplied range time index series share no temporal overlap.
        """
        # Define `series_2` so that it starts after `series_1` ends:
        series_1 = linear_timeseries(start=1, end=11, freq=2)
        series_2 = linear_timeseries(start=series_1.end_time() + 1, length=5, freq=3)
        self.assertEqual(get_shared_times(series_1, series_2), None)
        self.assertEqual(get_shared_times(series_1, series_1, series_2), None)
        self.assertEqual(get_shared_times(series_1, series_2, series_2), None)
        self.assertEqual(get_shared_times(series_1, series_1, series_2, series_2), None)

    def test_shared_times_no_overlap_datetime_idx(self):
        """
        Tests that `get_shared_times` returns `None` when
        supplied datetime time index series share no temporal overlap.
        """
        # Define `series_2` so that it starts after `series_1` ends:
        series_1 = linear_timeseries(
            start=pd.Timestamp("1/1/2000"), end=pd.Timestamp("1/11/2000"), freq="2d"
        )
        series_2 = linear_timeseries(
            start=series_1.end_time() + pd.Timedelta(1, "d"), length=5, freq="3d"
        )
        self.assertEqual(get_shared_times(series_1, series_2), None)
        self.assertEqual(get_shared_times(series_1, series_1, series_2), None)
        self.assertEqual(get_shared_times(series_1, series_2, series_2), None)
        self.assertEqual(get_shared_times(series_1, series_1, series_2, series_2), None)

    def test_shared_times_single_time_point_overlap_range_idx(self):
        """
        Tests that `get_shared_times` returns correct bounds when
        given range index series that overlap at a single time point.
        """
        # `series_1` and `series_2` only overlap at `series_1.end_time()`:
        series_1 = linear_timeseries(start=1, end=11, freq=2)
        series_2 = linear_timeseries(start=series_1.end_time(), length=5, freq=3)
        overlap_val = series_1.end_time()
        self.assertEqual(get_shared_times(series_1, series_2), overlap_val)
        self.assertEqual(get_shared_times(series_1, series_1, series_2), overlap_val)
        self.assertEqual(get_shared_times(series_1, series_2, series_2), overlap_val)
        self.assertEqual(
            get_shared_times(series_1, series_1, series_2, series_2), overlap_val
        )

    def test_shared_times_single_time_point_overlap_datetime_idx(self):
        """
        Tests that `get_shared_times` returns correct bounds when
        given datetime index series that overlap at a single time point.
        """
        # `series_1` and `series_2` only overlap at `series_1.end_time()`:
        series_1 = linear_timeseries(
            start=pd.Timestamp("1/1/2000"), end=pd.Timestamp("1/11/2000"), freq="2d"
        )
        series_2 = linear_timeseries(start=series_1.end_time(), length=5, freq="3d")
        overlap_val = series_1.end_time()
        self.assertEqual(get_shared_times(series_1, series_2), overlap_val)
        self.assertEqual(get_shared_times(series_1, series_1, series_2), overlap_val)
        self.assertEqual(get_shared_times(series_1, series_2, series_2), overlap_val)
        self.assertEqual(
            get_shared_times(series_1, series_1, series_2, series_2), overlap_val
        )

    def test_shared_times_identical_inputs_range_idx(self):
        """
        Tests that `get_shared_times` correctly handles case where
        multiple copies of same range index timeseries is passed;
        we expect that the unaltered time index of the series is returned.
        """
        series = linear_timeseries(start=0, length=5, freq=1)
        self.assertTrue(series.time_index.equals(get_shared_times(series)))
        self.assertTrue(series.time_index.equals(get_shared_times(series, series)))
        self.assertTrue(
            series.time_index.equals(get_shared_times(series, series, series))
        )

    def test_shared_times_identical_inputs_datetime_idx(self):
        """
        Tests that `get_shared_times` correctly handles case where
        multiple copies of same datetime index timeseries is passed;
        we expect that the unaltered time index of the series is returned.
        """
        series = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=5, freq="d")
        self.assertTrue(series.time_index.equals(get_shared_times(series)))
        self.assertTrue(series.time_index.equals(get_shared_times(series, series)))
        self.assertTrue(
            series.time_index.equals(get_shared_times(series, series, series))
        )

    def test_shared_times_unspecified_inputs(self):
        """
        Tests that `get_shared_times` correctly handles unspecified
        (i.e. `None` value) inputs. If `None` is passed with another
        series/time index, then `None` should be ignored and the time
        index of the other series should be returned. If only `None`
        values are passed, `None` should be returned.
        """
        # Pass `None` with series/time index input:
        series = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=5, freq="d")
        self.assertEqual(get_shared_times(None), None)
        self.assertTrue(series.time_index.equals(get_shared_times(series, None)))
        self.assertTrue(series.time_index.equals(get_shared_times(None, series, None)))
        self.assertTrue(
            series.time_index.equals(get_shared_times(None, series.time_index, None))
        )
        # Pass only `None` as input:
        self.assertEqual(get_shared_times(None), None)

    def test_shared_times_time_index_inputs(self):
        """
        Tests that `get_shared_times` can accept time index
        inputs instead of `TimeSeries` inputs; combinations
        of time index and `TimeSeries` inputs are also tested.
        """
        series_1 = linear_timeseries(start=0, end=10, freq=1)
        series_2 = linear_timeseries(start=0, end=20, freq=2)
        # `stop=10+1` since `stop` is exclusive
        intersection = pd.RangeIndex(
            start=series_2.start_time(), stop=series_1.end_time() + 1, step=2
        )
        self.assertTrue(
            intersection.equals(get_shared_times(series_1.time_index, series_2))
        )
        self.assertTrue(
            intersection.equals(get_shared_times(series_1, series_2.time_index))
        )
        self.assertTrue(
            intersection.equals(
                get_shared_times(series_1.time_index, series_2.time_index)
            )
        )

    def test_shared_times_empty_input(self):
        """
        Tests that `get_shared_times` returns `None` when
        given a non-`None` input with no timesteps.
        """
        series = linear_timeseries(start=0, length=0, freq=1)
        self.assertEqual(get_shared_times(series), None)
        self.assertEqual(get_shared_times(series.time_index), None)
        self.assertEqual(get_shared_times(series, series.time_index), None)

    def test_shared_times_different_time_index_types_error(self):
        """
        Tests that `get_shared_times` throws correct error when
        provided with series with different types of time indices.
        """
        series_1 = linear_timeseries(start=1, length=5, freq=1)
        series_2 = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=5, freq="d")
        with self.assertRaises(ValueError) as e:
            get_shared_times(series_1, series_2)
        self.assertEqual(
            (
                "Specified series and/or times must all "
                "have the same type of `time_index` (i.e. all "
                "`pd.RangeIndex` or all `pd.DatetimeIndex`)."
            ),
            str(e.exception),
        )



"""
Anomaly models base classes
"""

from abc import ABC, abstractmethod
from typing import Dict, Sequence, Union

from darts.ad.scorers.scorers import AnomalyScorer
from darts.ad.utils import (
    _to_list,
    eval_accuracy_from_scores,
    show_anomalies_from_scores,
)
from darts.logging import raise_if_not
from darts.timeseries import TimeSeries


class AnomalyModel(ABC):
    """Base class for all anomaly models."""

    def __init__(self, model, scorer):

        self.scorers = _to_list(scorer)

        raise_if_not(
            all([isinstance(s, AnomalyScorer) for s in self.scorers]),
            "all scorers must be of instance darts.ad.scorers.AnomalyScorer.",
        )

        self.scorers_are_trainable = any(s.trainable for s in self.scorers)
        self.univariate_scoring = any(s.univariate_scorer for s in self.scorers)

        self.model = model

    def _check_univariate(self, actual_anomalies):
        """Checks if `actual_anomalies` contains only univariate series, which
        is required if any of the scorers returns a univariate score.
        """

        if self.univariate_scoring:
            raise_if_not(
                all([s.width == 1 for s in actual_anomalies]),
                "Anomaly model contains scorer {} that will return".format(
                    [s.__str__() for s in self.scorers if s.univariate_scorer]
                )
                + " a univariate anomaly score series (width=1). Found a"
                + " multivariate `actual_anomalies`. The evaluation of the"
                + " accuracy cannot be computed. If applicable, think about"
                + " setting the scorer parameter `componenet_wise` to True.",
            )

    @abstractmethod
    def fit(
        self, series: Union[TimeSeries, Sequence[TimeSeries]]
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        pass

    @abstractmethod
    def score(
        self, series: Union[TimeSeries, Sequence[TimeSeries]]
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        pass

    @abstractmethod
    def eval_accuracy(
        self,
        actual_anomalies: Union[TimeSeries, Sequence[TimeSeries]],
        series: Union[TimeSeries, Sequence[TimeSeries]],
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        pass

    @abstractmethod
    def show_anomalies(self, series: TimeSeries):
        pass

    def _show_anomalies(
        self,
        series: TimeSeries,
        model_output: TimeSeries = None,
        anomaly_scores: Union[TimeSeries, Sequence[TimeSeries]] = None,
        names_of_scorers: Union[str, Sequence[str]] = None,
        actual_anomalies: TimeSeries = None,
        title: str = None,
        metric: str = None,
    ):
        """Internal function that plots the results of the anomaly model.
        Called by the function show_anomalies().
        """

        if title is None:
            title = f"Anomaly results ({self.model.__class__.__name__})"

        if names_of_scorers is None:
            names_of_scorers = [s.__str__() for s in self.scorers]

        list_window = [s.window for s in self.scorers]

        return show_anomalies_from_scores(
            series,
            model_output=model_output,
            anomaly_scores=anomaly_scores,
            window=list_window,
            names_of_scorers=names_of_scorers,
            actual_anomalies=actual_anomalies,
            title=title,
            metric=metric,
        )

    def _eval_accuracy_from_scores(
        self,
        list_actual_anomalies: Sequence[TimeSeries],
        list_anomaly_scores: Sequence[TimeSeries],
        metric: str,
    ) -> Union[Sequence[Dict[str, float]], Sequence[Dict[str, Sequence[float]]]]:
        """Internal function that computes the accuracy of the anomaly scores
        computed by the model. Called by the function eval_accuracy().
        """
        windows = [s.window for s in self.scorers]

        # create a list of unique names for each scorer that
        # will be used as keys for the dictionary containing
        # the accuracy of each scorer.
        name_scorers = []
        for scorer in self.scorers:
            name = scorer.__str__() + "_w=" + str(scorer.window)

            if name in name_scorers:
                i = 1
                new_name = name + "_" + str(i)
                while new_name in name_scorers:
                    i = i + 1
                    new_name = name + "_" + str(i)
                name = new_name

            name_scorers.append(name)

        acc = []
        for anomalies, scores in zip(list_actual_anomalies, list_anomaly_scores):
            acc.append(
                eval_accuracy_from_scores(
                    actual_anomalies=anomalies,
                    anomaly_score=scores,
                    window=windows,
                    metric=metric,
                )
            )

        return [dict(zip(name_scorers, scorer_values)) for scorer_values in acc]



"""
Anomaly Detectors
-----------------

Detectors provide binary anomaly classification on time series.
They can typically be used to transform anomaly scores time series into binary anomaly time series.

Some detectors are trainable. For instance, ``QuantileDetector`` emits a binary anomaly for
every time step where the observed value(s) are beyond the quantile(s) observed
on the training series.

The main functions are ``fit()`` (for the trainable detectors), ``detect()`` and ``eval_accuracy()``.

``fit()`` trains the detector over the history of one or multiple time series. It can
for instance be called on series containing anomaly scores (or even raw values) during normal times.
The function ``detect()`` takes an anomaly score time series as input, and applies the detector
to obtain binary predictions. The function ``eval_accuracy()`` returns the accuracy metric
("accuracy", "precision", "recall" or "f1") between a binary prediction time series and some known
binary ground truth time series indicating the presence of anomalies.
"""

from .quantile_detector import QuantileDetector
from .threshold_detector import ThresholdDetector



"""
Scaler
------
"""

from copy import deepcopy
from typing import Any, Mapping, Sequence, Union

import numpy as np
from sklearn.preprocessing import MinMaxScaler

from darts.logging import get_logger, raise_log
from darts.timeseries import TimeSeries

from .fittable_data_transformer import FittableDataTransformer
from .invertible_data_transformer import InvertibleDataTransformer

logger = get_logger(__name__)


class Scaler(FittableDataTransformer, InvertibleDataTransformer):
    def __init__(
        self,
        scaler=None,
        name="Scaler",
        global_fit: bool = False,
        n_jobs: int = 1,
        verbose: bool = False,
    ):
        """Generic wrapper class for using scalers on time series.

        The underlying `scaler` has to implement the ``fit()``, ``transform()`` and
        ``inverse_transform()`` methods (typically from scikit-learn).

        When the scaler is applied on multivariate series, the scaling is done per-component.
        When the series are stochastic, the scaling is done across all samples (for each given component).
        The transformation is applied independently for each dimension (component) of the time series,
        effectively merging all samples of a component in order to compute the transform.

        Notes
        -----
        The scaler will not scale the series' static covariates. This has to be done either before constructing the
        series, or later on by extracting the covariates, transforming the values and then reapplying them to the
        series. For this, see TimeSeries properties `TimeSeries.static_covariates` and method
        `TimeSeries.with_static_covariates()`

        Parameters
        ----------
        scaler
            The scaler to transform the data with. It must provide ``fit()``,
            ``transform()`` and ``inverse_transform()`` methods.
            Default: :class:`sklearn.preprocessing.MinMaxScaler(feature_range=(0, 1))`; this will scale all
            the values of a time series between 0 and 1.
        name
            A specific name for the scaler
        global_fit
            Optionally, whether all of the `TimeSeries` passed to the `fit()` method should be used to fit
            a *single* set of parameters, or if a different set of parameters should be independently fitted
            to each provided `TimeSeries`. If `True`, then a `Sequence[TimeSeries]` is passed to `ts_fit`
            and a single set of parameters is fitted using all of the provided `TimeSeries`. If `False`, then
            each `TimeSeries` is individually passed to `ts_fit`, and a different set of fitted parameters
            if yielded for each of these fitting operations. See `FittableDataTransformer` documentation for
            further details.
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a ``Sequence[TimeSeries]`` is
            passed as input to a method, parallelising operations regarding different ``TimeSeries``. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        verbose
            Optionally, whether to print operations progress

        Notes
        -----
        In case the :class:`Scaler` is applied to multiple ``TimeSeries`` objects, a deep-copy of the
        chosen scaler will be instantiated, fitted, and stored, for each ``TimeSeries``.

        Examples
        --------
        >>> from darts.datasets import AirPassengersDataset
        >>> from sklearn.preprocessing import MinMaxScaler
        >>> from darts.dataprocessing.transformers import Scaler
        >>> series = AirPassengersDataset().load()
        >>> scaler = MinMaxScaler(feature_range=(-1, 1))
        >>> transformer = Scaler(scaler)
        >>> series_transformed = transformer.fit_transform(series)
        >>> print(min(series_transformed.values()))
        [-1.]
        >>> print(max(series_transformed.values()))
        [2.]
        """

        if scaler is None:
            scaler = MinMaxScaler(feature_range=(0, 1))

        if (
            not callable(getattr(scaler, "fit", None))
            or not callable(getattr(scaler, "transform", None))
            or not callable(getattr(scaler, "inverse_transform", None))
        ):
            raise_log(
                ValueError(
                    "The provided transformer object must have fit(), transform() and inverse_transform() methods"
                ),
                logger,
            )
        # Define fixed params (i.e. attributes defined before calling `super().__init__`):
        self.transformer = scaler
        super().__init__(
            name=name, n_jobs=n_jobs, verbose=verbose, global_fit=global_fit
        )

    @staticmethod
    def ts_transform(
        series: TimeSeries, params: Mapping[str, Any], **kwargs
    ) -> TimeSeries:

        transformer = params["fitted"]

        tr_out = transformer.transform(Scaler.stack_samples(series))

        transformed_vals = Scaler.unstack_samples(tr_out, series=series)

        return series.with_values(transformed_vals)

    @staticmethod
    def ts_inverse_transform(
        series: TimeSeries, params: Mapping[str, Any], *args, **kwargs
    ) -> TimeSeries:
        transformer = params["fitted"]

        tr_out = transformer.inverse_transform(Scaler.stack_samples(series))
        inv_transformed_vals = Scaler.unstack_samples(tr_out, series=series)

        return series.with_values(inv_transformed_vals)

    @staticmethod
    def ts_fit(
        series: Union[TimeSeries, Sequence[TimeSeries]],
        params: Mapping[str, Any],
        *args,
        **kwargs
    ) -> Any:
        transformer = deepcopy(params["fixed"]["transformer"])
        # If `global_fit` is `True`, then `series` will be ` Sequence[TimeSeries]`;
        # otherwise, `series` is a single `TimeSeries`:
        if isinstance(series, TimeSeries):
            series = [series]
        vals = np.concatenate([Scaler.stack_samples(ts) for ts in series], axis=0)
        scaler = transformer.fit(vals)
        return scaler



"""
Ensemble scikit-learn aggregator
--------------------------------

Aggregator wrapped around the Ensemble model of sklearn.
`sklearn https://scikit-learn.org/stable/modules/ensemble.html`_.
"""

from typing import Sequence

import numpy as np
from sklearn.ensemble import BaseEnsemble

from darts import TimeSeries
from darts.ad.aggregators.aggregators import FittableAggregator
from darts.logging import raise_if_not


class EnsembleSklearnAggregator(FittableAggregator):
    def __init__(self, model) -> None:

        raise_if_not(
            isinstance(model, BaseEnsemble),
            f"Scorer is expecting a model of type BaseEnsemble (from sklearn ensemble), \
            found type {type(model)}.",
        )

        self.model = model
        super().__init__()

    def __str__(self):
        return "EnsembleSklearnAggregator: {}".format(
            self.model.__str__().split("(")[0]
        )

    def _fit_core(
        self,
        actual_anomalies: Sequence[TimeSeries],
        series: Sequence[TimeSeries],
    ):

        X = np.concatenate(
            [s.all_values(copy=False).reshape(len(s), -1) for s in series],
            axis=0,
        )

        y = np.concatenate(
            [s.all_values(copy=False).reshape(len(s)) for s in actual_anomalies],
            axis=0,
        )

        self.model.fit(y=y, X=X)
        return self

    def _predict_core(self, series: Sequence[TimeSeries]) -> Sequence[TimeSeries]:

        return [
            TimeSeries.from_times_and_values(
                s.time_index,
                self.model.predict((s).all_values(copy=False).reshape(len(s), -1)),
            )
            for s in series
        ]



"""
Baseline Models
---------------

A collection of simple benchmark models for univariate series.
"""

from typing import List, Optional, Sequence, Union

import numpy as np

from darts.logging import get_logger, raise_if_not
from darts.models.forecasting.ensemble_model import EnsembleModel
from darts.models.forecasting.forecasting_model import (
    GlobalForecastingModel,
    LocalForecastingModel,
)
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class NaiveMean(LocalForecastingModel):
    def __init__(self):
        """Naive Mean Model

        This model has no parameter, and always predicts the
        mean value of the training series.
        """
        super().__init__()
        self.mean_val = None

    def fit(self, series: TimeSeries):
        super().fit(series)

        self.mean_val = np.mean(series.values(copy=False), axis=0)
        return self

    def predict(self, n: int, num_samples: int = 1, verbose: bool = False):
        super().predict(n, num_samples)
        forecast = np.tile(self.mean_val, (n, 1))
        return self._build_forecast_series(forecast)


class NaiveSeasonal(LocalForecastingModel):
    def __init__(self, K: int = 1):
        """Naive Seasonal Model

        This model always predicts the value of `K` time steps ago.
        When `K=1`, this model predicts the last value of the training set.
        When `K>1`, it repeats the last `K` values of the training set.

        Parameters
        ----------
        K
            the number of last time steps of the training set to repeat
        """
        super().__init__()
        self.last_k_vals = None
        self.K = K

    @property
    def min_train_series_length(self):
        return max(self.K, 3)

    def fit(self, series: TimeSeries):
        super().fit(series)

        raise_if_not(
            len(series) >= self.K,
            f"The time series requires at least K={self.K} points",
            logger,
        )
        self.last_k_vals = series.values(copy=False)[-self.K :, :]
        return self

    def predict(self, n: int, num_samples: int = 1, verbose: bool = False):
        super().predict(n, num_samples)
        forecast = np.array([self.last_k_vals[i % self.K, :] for i in range(n)])
        return self._build_forecast_series(forecast)


class NaiveDrift(LocalForecastingModel):
    def __init__(self):
        """Naive Drift Model

        This model fits a line between the first and last point of the training series,
        and extends it in the future. For a training series of length :math:`T`, we have:

        .. math:: \\hat{y}_{T+h} = y_T + h\\left( \\frac{y_T - y_1}{T - 1} \\right)
        """
        super().__init__()

    def fit(self, series: TimeSeries):
        super().fit(series)
        assert series.n_samples == 1, "This model expects deterministic time series"

        series = self.training_series
        return self

    def predict(self, n: int, num_samples: int = 1, verbose: bool = False):
        super().predict(n, num_samples)
        first, last = (
            self.training_series.first_values(),
            self.training_series.last_values(),
        )
        slope = (last - first) / (len(self.training_series) - 1)
        last_value = last + slope * n
        forecast = np.linspace(last, last_value, num=n + 1)[1:]
        return self._build_forecast_series(forecast)


class NaiveMovingAverage(LocalForecastingModel):
    def __init__(self, input_chunk_length: int = 1):
        """Naive Moving Average Model

        This model forecasts using an auto-regressive moving average (ARMA).

        Parameters
        ----------
        input_chunk_length
            The size of the sliding window used to calculate the moving average
        """
        super().__init__()
        self.input_chunk_length = input_chunk_length
        self.rolling_window = None

    @property
    def min_train_series_length(self):
        return self.input_chunk_length

    def __str__(self):
        return f"NaiveMovingAverage({self.input_chunk_length})"

    def fit(self, series: TimeSeries):
        super().fit(series)
        raise_if_not(
            series.is_deterministic,
            "This model expects deterministic time series",
            logger,
        )

        self.rolling_window = series[-self.input_chunk_length :].values(copy=False)
        return self

    def predict(self, n: int, num_samples: int = 1, verbose: bool = False):
        super().predict(n, num_samples)

        predictions_with_observations = np.concatenate(
            (self.rolling_window, np.zeros(shape=(n, self.rolling_window.shape[1]))),
            axis=0,
        )
        rolling_sum = sum(self.rolling_window)

        chunk_length = self.input_chunk_length
        for i in range(chunk_length, chunk_length + n):
            prediction = rolling_sum / chunk_length
            predictions_with_observations[i] = prediction
            lost_value = predictions_with_observations[i - chunk_length]
            rolling_sum += prediction - lost_value
        return self._build_forecast_series(predictions_with_observations[-n:])


class NaiveEnsembleModel(EnsembleModel):
    def __init__(
        self, models: Union[List[LocalForecastingModel], List[GlobalForecastingModel]]
    ):
        """Naive combination model

        Naive implementation of `EnsembleModel`
        Returns the average of all predictions of the constituent models
        """
        super().__init__(models)

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    ):
        super().fit(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )
        for model in self.models:
            if self.is_global_ensemble:
                kwargs = dict(series=series)
                if model.supports_past_covariates:
                    kwargs["past_covariates"] = past_covariates
                if model.supports_future_covariates:
                    kwargs["future_covariates"] = future_covariates
                model.fit(**kwargs)
            else:
                model.fit(series=series)

        return self

    def ensemble(
        self,
        predictions: Union[TimeSeries, Sequence[TimeSeries]],
        series: Optional[Sequence[TimeSeries]] = None,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        def take_average(prediction: TimeSeries) -> TimeSeries:
            series = prediction.pd_dataframe(copy=False).sum(axis=1) / len(self.models)
            series.name = prediction.components[0]
            return TimeSeries.from_series(series)

        if isinstance(predictions, Sequence):
            return [take_average(p) for p in predictions]
        else:
            return take_average(predictions)



"""
Box-Cox Transformer
-------------------
"""

from typing import Any, Mapping, Optional, Sequence, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np
import pandas as pd
from scipy.special import inv_boxcox
from scipy.stats import boxcox, boxcox_normmax

from darts.logging import get_logger, raise_if
from darts.timeseries import TimeSeries

from .fittable_data_transformer import FittableDataTransformer
from .invertible_data_transformer import InvertibleDataTransformer

logger = get_logger(__name__)


class BoxCox(FittableDataTransformer, InvertibleDataTransformer):
    def __init__(
        self,
        name: str = "BoxCox",
        lmbda: Optional[
            Union[float, Sequence[float], Sequence[Sequence[float]]]
        ] = None,
        optim_method: Literal["mle", "pearsonr"] = "mle",
        global_fit: bool = False,
        n_jobs: int = 1,
        verbose: bool = False,
    ):
        """Box-Cox data transformer.

        See [1]_ for more information about Box-Cox transforms.

        The transformation is applied independently for each dimension (component) of the time series.
        For stochastic series, it is done jointly over all samples, effectively merging all samples of
        a component in order to compute the transform.

        Notes
        -----
        The scaler will not scale the series' static covariates. This has to be done either before constructing the
        series, or later on by extracting the covariates, transforming the values and then reapplying them to the
        series. For this, see TimeSeries properties `TimeSeries.static_covariates` and method
        `TimeSeries.with_static_covariates()`

        Parameters
        ----------
        name
            A specific name for the transformer
        lmbda
            The parameter :math:`\\lambda` of the Box-Cox transform. If a single float is given, the same
            :math:`\\lambda` value will be used for all dimensions of the series, for all the series.
            If a sequence is given, there is one value per component in the series. If a sequence of sequence
            is given, there is one value per component for all series.
            If `None` given, will automatically find an optimal value of :math:`\\lambda` (for each dimension
            of the time series, for each time series) using :func:`scipy.stats.boxcox_normmax`
            with ``method=optim_method``.
        optim_method
            Specifies which method to use to find an optimal value for the lmbda parameter.
            Either 'mle' or 'pearsonr'. Ignored if `lmbda` is not `None`.
        global_fit
            Optionally, whether all of the `TimeSeries` passed to the `fit()` method should be used to fit
            a *single* set of parameters, or if a different set of parameters should be independently fitted
            to each provided `TimeSeries`. If `True`, then a `Sequence[TimeSeries]` is passed to `ts_fit`
            and a single set of parameters is fitted using all of the provided `TimeSeries`. If `False`, then
            each `TimeSeries` is individually passed to `ts_fit`, and a different set of fitted parameters
            if yielded for each of these fitting operations. See `FittableDataTransformer` documentation for
            further details.
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a ``Sequence[TimeSeries]`` is
            passed as input, parallelising operations regarding different ``TimeSeries``. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        verbose
            Whether to print operations progress

        Examples
        --------
        >>> from darts.datasets import AirPassengersDataset
        >>> from darts.dataprocessing.transformers import BoxCox
        >>> series = AirPassengersDataset().load()
        >>> transformer = BoxCox(lmbda=0.2)
        >>> series_transformed = transformer.fit_transform(series)
        >>> print(series_transformed.head())
        <TimeSeries (DataArray) (Month: 5, component: 1, sample: 1)>
        array([[[7.84735157]],
            [[7.98214351]],
            [[8.2765364 ]],
            [[8.21563229]],
            [[8.04749318]]])
        Coordinates:
        * Month      (Month) datetime64[ns] 1949-01-01 1949-02-01 ... 1949-05-01
        * component  (component) object '#Passengers'
        Dimensions without coordinates: sample

        References
        ----------
        .. [1] https://otexts.com/fpp2/transformations.html#mathematical-transformations
        """
        raise_if(
            not isinstance(optim_method, str)
            or optim_method not in ["mle", "pearsonr"],
            "optim_method parameter must be either 'mle' or 'pearsonr'",
            logger,
        )

        # Define fixed params (i.e. attributes defined before calling `super().__init__`):
        self._lmbda = lmbda
        self._optim_method = optim_method
        if isinstance(lmbda, Sequence) and isinstance(lmbda[0], Sequence):
            parallel_params = ("_lmbda",)
        else:
            parallel_params = False

        super().__init__(
            name=name,
            n_jobs=n_jobs,
            verbose=verbose,
            parallel_params=parallel_params,
            mask_components=True,
            global_fit=global_fit,
        )

    @staticmethod
    def ts_fit(
        series: Union[TimeSeries, Sequence[TimeSeries]],
        params: Mapping[str, Any],
        *args,
        **kwargs
    ) -> Union[Sequence[float], pd.core.series.Series]:
        lmbda, method = params["fixed"]["_lmbda"], params["fixed"]["_optim_method"]
        # If `global_fit` is `True`, then `series` will be ` Sequence[TimeSeries]`;
        # otherwise, `series` is a single `TimeSeries`:
        if isinstance(series, TimeSeries):
            series = [series]
        if lmbda is None:
            # Compute optimal lmbda for each dimension of the time series. In this case, the return type is
            # an ndarray and not a Sequence
            vals = np.concatenate([BoxCox.stack_samples(ts) for ts in series], axis=0)
            lmbda = np.apply_along_axis(boxcox_normmax, axis=0, arr=vals, method=method)

        elif isinstance(lmbda, Sequence):
            raise_if(
                len(lmbda) != series[0].width,
                "lmbda should have one value per dimension (ie. column or variable) of the time series",
                logger,
            )
        else:
            # Replicate lmbda to match dimensions of the time series
            lmbda = [lmbda] * series[0].width

        return lmbda

    @staticmethod
    def ts_transform(
        series: TimeSeries, params: Mapping[str, Any], **kwargs
    ) -> TimeSeries:

        lmbda = params["fitted"]

        vals = BoxCox.stack_samples(series)
        transformed_vals = np.stack(
            [boxcox(vals[:, i], lmbda=lmbda[i]) for i in range(series.width)], axis=1
        )
        transformed_vals = BoxCox.unstack_samples(transformed_vals, series=series)
        return series.with_values(transformed_vals)

    @staticmethod
    def ts_inverse_transform(
        series: TimeSeries, params: Mapping[str, Any], **kwargs
    ) -> TimeSeries:

        lmbda = params["fitted"]

        vals = BoxCox.stack_samples(series)
        inv_transformed_vals = np.stack(
            [inv_boxcox(vals[:, i], lmbda[i]) for i in range(series.width)], axis=1
        )
        inv_transformed_vals = BoxCox.unstack_samples(
            inv_transformed_vals, series=series
        )
        return series.with_values(inv_transformed_vals)



"""
StatsForecastAutoCES
-----------
"""

from statsforecast.models import AutoCES as SFAutoCES

from darts import TimeSeries
from darts.models.forecasting.forecasting_model import LocalForecastingModel


class StatsForecastAutoCES(LocalForecastingModel):
    def __init__(self, *autoces_args, **autoces_kwargs):
        """Auto-CES based on `Statsforecasts package
        <https://github.com/Nixtla/statsforecast>`_.

        Automatically selects the best Complex Exponential Smoothing model using an information criterion.
        <https://onlinelibrary.wiley.com/doi/full/10.1002/nav.22074>

        We refer to the `statsforecast AutoCES documentation
        <https://nixtla.github.io/statsforecast/models.html#autoces>`_
        for the documentation of the arguments.

        Parameters
        ----------
        autoces_args
            Positional arguments for ``statsforecasts.models.AutoCES``.
        autoces_kwargs
            Keyword arguments for ``statsforecasts.models.AutoCES``.

            ..

        Examples
        --------
        >>> from darts.models import StatsForecastAutoCES
        >>> from darts.datasets import AirPassengersDataset
        >>> series = AirPassengersDataset().load()
        >>> model = StatsForecastAutoCES(season_length=12)
        >>> model.fit(series[:-36])
        >>> pred = model.predict(36, num_samples=100)
        """
        super().__init__()
        self.model = SFAutoCES(*autoces_args, **autoces_kwargs)

    def fit(self, series: TimeSeries):
        super().fit(series)
        self._assert_univariate(series)
        series = self.training_series
        self.model.fit(
            series.values(copy=False).flatten(),
        )
        return self

    def predict(
        self,
        n: int,
        num_samples: int = 1,
        verbose: bool = False,
    ):
        super().predict(n, num_samples)
        forecast_dict = self.model.predict(
            h=n,
        )

        mu = forecast_dict["mean"]

        return self._build_forecast_series(mu)

    @property
    def min_train_series_length(self) -> int:
        return 10

    def _supports_range_index(self) -> bool:
        return True

    def _is_probabilistic(self) -> bool:
        return False



"""
Kalman Filter Forecaster
------------------------

A model producing stochastic forecasts based on the Kalman filter.
The filter is first optionally fitted on the series (using the N4SID
identification algorithm), and then run on future time steps in order
to obtain forecasts.

This implementation accepts an optional control signal (future covariates).
"""

from typing import Optional

import numpy as np
from nfoursid.kalman import Kalman

from darts.logging import get_logger
from darts.models.filtering.kalman_filter import KalmanFilter
from darts.models.forecasting.forecasting_model import (
    TransferableFutureCovariatesLocalForecastingModel,
)
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class KalmanForecaster(TransferableFutureCovariatesLocalForecastingModel):
    def __init__(
        self,
        dim_x: int = 1,
        kf: Optional[Kalman] = None,
        add_encoders: Optional[dict] = None,
    ):
        """Kalman filter Forecaster

        This model uses a Kalman filter to produce forecasts. It uses a
        :class:`darts.models.filtering.kalman_filter.KalmanFilter` object
        and treats future values as missing values.

        The model can optionally receive a :class:`nfoursid.kalman.Kalman`
        object specifying the Kalman filter, or, if not specified, the filter
        will be trained using the N4SID system identification algorithm.

        Parameters
        ----------
        dim_x : int
            Size of the Kalman filter state vector.
        kf : nfoursid.kalman.Kalman
            Optionally, an instance of `nfoursid.kalman.Kalman`.
            If this is provided, the parameter dim_x is ignored. This instance will be copied for every
            call to `predict()`, so the state is not carried over from one time series to another across several
            calls to `predict()`.
            The various dimensionalities of the filter must match those of the `TimeSeries` used when
            calling `predict()`.
            If this is specified, it is still necessary to call `fit()` before calling `predict()`,
            although this will have no effect on the Kalman filter.
        add_encoders
            A large number of future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'future': ['relative']},
                    'custom': {'future': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
            ..
        """
        super().__init__(add_encoders=add_encoders)
        self.dim_x = dim_x
        self.kf = kf
        self.darts_kf = KalmanFilter(dim_x, kf)

    def _fit(self, series: TimeSeries, future_covariates: Optional[TimeSeries] = None):

        super()._fit(series, future_covariates)
        if self.kf is None:
            self.darts_kf.fit(series=series, covariates=future_covariates)

        return self

    def predict(
        self,
        n: int,
        series: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        **kwargs,
    ) -> TimeSeries:
        # we override `predict()` to pass a non-None `series`, so that historic_future_covariates
        # will be passed to `_predict()`
        series = series if series is not None else self.training_series
        return super().predict(n, series, future_covariates, num_samples, **kwargs)

    def _predict(
        self,
        n: int,
        series: Optional[TimeSeries] = None,
        historic_future_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        verbose: bool = False,
    ) -> TimeSeries:

        super()._predict(
            n, series, historic_future_covariates, future_covariates, num_samples
        )
        time_index = self._generate_new_dates(n, input_series=series)
        placeholder_vals = np.zeros((n, self.training_series.width)) * np.nan
        series_future = TimeSeries.from_times_and_values(
            time_index,
            placeholder_vals,
            columns=self.training_series.columns,
            static_covariates=self.training_series.static_covariates,
            hierarchy=self.training_series.hierarchy,
        )

        series = series.append(series_future)
        if historic_future_covariates is not None:
            future_covariates = historic_future_covariates.append(future_covariates)

        filtered_series = self.darts_kf.filter(
            series=series, covariates=future_covariates, num_samples=num_samples
        )

        return filtered_series[-n:]

    def _is_probabilistic(self) -> bool:
        return True



import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.tests.base_test_class import DartsBaseTestClass
from darts.tests.test_timeseries import TimeSeriesTestCase


class TimeSeriesMultivariateTestCase(DartsBaseTestClass):

    times1 = pd.date_range("20130101", "20130110")
    times2 = pd.date_range("20130206", "20130215")
    dataframe1 = pd.DataFrame(
        {"0": range(10), "1": range(5, 15), "2": range(10, 20)}, index=times1
    )
    dataframe2 = pd.DataFrame(
        {"0": np.arange(1, 11), "1": np.arange(1, 11) * 3, "2": np.arange(1, 11) * 5},
        index=times1,
    )
    dataframe3 = pd.DataFrame(
        {
            "0": np.arange(1, 11),
            "1": np.arange(11, 21),
        },
        index=times2,
    )
    dataframe4 = pd.DataFrame(
        {
            "0": [1, 1, np.nan, 1, 1, 1, 1, 1, 1, 1],
            "1": [1, 1, np.nan, 1, 1, np.nan, np.nan, 1, 1, 1],
            "2": [1, 1, np.nan, 1, 1, np.nan, np.nan, np.nan, np.nan, 1],
        },
        index=times2,
    )
    series1 = TimeSeries.from_dataframe(dataframe1)
    series2 = TimeSeries.from_dataframe(dataframe2)
    series3 = TimeSeries.from_dataframe(dataframe3)
    series4 = TimeSeries.from_dataframe(dataframe4)

    def test_creation(self):
        series_test = TimeSeries.from_dataframe(self.dataframe1)
        self.assertTrue(
            np.all(series_test.pd_dataframe().values == (self.dataframe1.values))
        )

        # Series cannot be lower than three without passing frequency as argument to constructor
        with self.assertRaises(ValueError):
            TimeSeries(self.dataframe1.iloc[:2, :])
        TimeSeries.from_dataframe(self.dataframe1.iloc[:2, :], freq="D")

    def test_eq(self):
        seriesA = TimeSeries.from_dataframe(self.dataframe1)
        self.assertTrue(self.series1 == seriesA)
        self.assertFalse(self.series1 != seriesA)

        # with different dates
        dataframeB = self.dataframe1.copy()
        dataframeB.index = pd.date_range("20130102", "20130111")
        seriesB = TimeSeries.from_dataframe(dataframeB)
        self.assertFalse(self.series1 == seriesB)

        # with one different value
        dataframeC = self.dataframe1.copy()
        dataframeC.iloc[2, 2] = 0
        seriesC = TimeSeries.from_dataframe(dataframeC)
        self.assertFalse(self.series1 == seriesC)

    def test_rescale(self):
        with self.assertRaises(ValueError):
            self.series1.rescale_with_value(1)

        seriesA = self.series2.rescale_with_value(0)
        self.assertTrue(np.all(seriesA.values() == 0).all())

        seriesB = self.series2.rescale_with_value(1)
        self.assertEqual(
            seriesB,
            TimeSeries.from_dataframe(
                pd.DataFrame(
                    {
                        "0": np.arange(1, 11),
                        "1": np.arange(1, 11),
                        "2": np.arange(1, 11),
                    },
                    index=self.dataframe2.index,
                ).astype(float)
            ),
        )

    def test_slice(self):
        TimeSeriesTestCase.helper_test_slice(self, self.series1)

    def test_split(self):
        TimeSeriesTestCase.helper_test_split(self, self.series1)

    def test_drop(self):
        TimeSeriesTestCase.helper_test_drop(self, self.series1)

    def test_intersect(self):
        TimeSeriesTestCase.helper_test_intersect(self, self.series1)

    def test_shift(self):
        TimeSeriesTestCase.helper_test_shift(self, self.series1)

    def test_append(self):
        TimeSeriesTestCase.helper_test_append(self, self.series1)

    def test_append_values(self):
        TimeSeriesTestCase.helper_test_append_values(self, self.series1)

    def test_prepend(self):
        TimeSeriesTestCase.helper_test_prepend(self, self.series1)

    def test_prepend_values(self):
        TimeSeriesTestCase.helper_test_prepend_values(self, self.series1)

    def test_strip(self):
        dataframe1 = pd.DataFrame(
            {
                "0": 2 * [np.nan] + list(range(7)) + [np.nan],
                "1": [np.nan] + list(range(7)) + 2 * [np.nan],
            },
            index=self.times1,
        )
        series1 = TimeSeries.from_dataframe(dataframe1)

        self.assertTrue((series1.strip().time_index == self.times1[1:-1]).all())

    """
    Testing new multivariate methods.
    """

    def test_stack(self):
        with self.assertRaises(ValueError):
            self.series1.stack(self.series3)
        seriesA = self.series1.stack(self.series2)
        dataframeA = pd.concat([self.dataframe1, self.dataframe2], axis=1)
        dataframeA.columns = [
            "0",
            "1",
            "2",
            "0_1",
            "1_1",
            "2_1",
        ]  # the names to expect after stacking
        self.assertTrue((seriesA.pd_dataframe() == dataframeA).all().all())
        self.assertEqual(
            seriesA.values().shape,
            (
                len(self.dataframe1),
                len(self.dataframe1.columns) + len(self.dataframe2.columns),
            ),
        )

    def test_univariate_component(self):
        with self.assertRaises(IndexError):
            self.series1.univariate_component(-5)
        with self.assertRaises(IndexError):
            self.series1.univariate_component(3)
        seriesA = self.series1.univariate_component(1)
        self.assertTrue(
            seriesA
            == TimeSeries.from_times_and_values(
                self.times1, range(5, 15), columns=["1"]
            )
        )
        seriesB = (
            self.series1.univariate_component(0)
            .stack(seriesA)
            .stack(self.series1.univariate_component(2))
        )
        self.assertTrue(self.series1 == seriesB)

    def test_add_datetime_attribute(self):
        seriesA = self.series1.add_datetime_attribute("day")
        self.assertEqual(seriesA.width, self.series1.width + 1)
        self.assertTrue(
            set(seriesA.pd_dataframe().iloc[:, seriesA.width - 1].values.flatten())
            == set(range(1, 11))
        )
        seriesB = self.series3.add_datetime_attribute("day", True)
        self.assertEqual(seriesB.width, self.series3.width + 31)
        self.assertEqual(
            set(seriesB.pd_dataframe().iloc[:, self.series3.width :].values.flatten()),
            {0, 1},
        )
        seriesC = self.series1.add_datetime_attribute("month", True)
        self.assertEqual(seriesC.width, self.series1.width + 12)
        seriesD = TimeSeries.from_times_and_values(
            pd.date_range("20130206", "20130430"), range(84)
        )
        seriesD = seriesD.add_datetime_attribute("month", True)
        self.assertEqual(seriesD.width, 13)
        self.assertEqual(sum(seriesD.values().flatten()), sum(range(84)) + 84)
        self.assertEqual(sum(seriesD.values()[:, 1 + 3]), 30)
        self.assertEqual(sum(seriesD.values()[:, 1 + 1]), 23)

        # test cyclic
        times_month = pd.date_range("20130101", "20140610")

        seriesE = TimeSeries.from_times_and_values(
            times_month, np.repeat(0.1, len(times_month))
        )
        seriesF = seriesE.add_datetime_attribute("day", cyclic=True)

        values_sin = seriesF.values()[:, 1]
        values_cos = seriesF.values()[:, 2]

        self.assertTrue(
            np.allclose(np.add(np.square(values_sin), np.square(values_cos)), 1)
        )

        df = seriesF.pd_dataframe()
        df = df[df.index.day == 1]
        self.assertTrue(np.allclose(df["day_sin"].values, 0.2, atol=0.03))
        self.assertTrue(np.allclose(df["day_cos"].values, 0.97, atol=0.03))

    def test_add_holidays(self):
        times = pd.date_range(start=pd.Timestamp("20201201"), periods=30, freq="D")
        seriesA = TimeSeries.from_times_and_values(times, range(len(times)))

        # testing for christmas and non-holiday in US
        seriesA = seriesA.add_holidays("US")
        last_column = seriesA.pd_dataframe().iloc[:, seriesA.width - 1]
        self.assertEqual(last_column.at[pd.Timestamp("20201225")], 1)
        self.assertEqual(last_column.at[pd.Timestamp("20201210")], 0)
        self.assertEqual(last_column.at[pd.Timestamp("20201226")], 0)

        # testing for christmas and non-holiday in PL
        seriesA = seriesA.add_holidays("PL")
        last_column = seriesA.pd_dataframe().iloc[:, seriesA.width - 1]
        self.assertEqual(last_column.at[pd.Timestamp("20201225")], 1)
        self.assertEqual(last_column.at[pd.Timestamp("20201210")], 0)
        self.assertEqual(last_column.at[pd.Timestamp("20201226")], 1)
        self.assertEqual(seriesA.width, 3)

        # testing hourly time series
        times = pd.date_range(start=pd.Timestamp("20201224"), periods=50, freq="H")
        seriesB = TimeSeries.from_times_and_values(times, range(len(times)))
        seriesB = seriesB.add_holidays("US")
        last_column = seriesB.pd_dataframe().iloc[:, seriesB.width - 1]
        self.assertEqual(last_column.at[pd.Timestamp("2020-12-25 01:00:00")], 1)
        self.assertEqual(last_column.at[pd.Timestamp("2020-12-24 23:00:00")], 0)

    def test_assert_univariate(self):
        with self.assertRaises(AssertionError):
            self.series1._assert_univariate()
        self.series1.univariate_component(0)._assert_univariate()

    def test_first_last_values(self):
        self.assertEqual(self.series1.first_values().tolist(), [0, 5, 10])
        self.assertEqual(self.series3.last_values().tolist(), [10, 20])
        self.assertEqual(
            self.series1.univariate_component(1).first_values().tolist(), [5]
        )
        self.assertEqual(
            self.series3.univariate_component(1).last_values().tolist(), [20]
        )

    def test_drop_column(self):
        # testing dropping a single column
        seriesA = self.series1.drop_columns("0")
        self.assertNotIn("0", seriesA.columns.values)
        self.assertEqual(seriesA.columns.tolist(), ["1", "2"])
        self.assertEqual(len(seriesA.columns), 2)

        # testing dropping multiple columns
        seriesB = self.series1.drop_columns(["0", "1"])
        self.assertIn("2", seriesB.columns.values)
        self.assertEqual(len(seriesB.columns), 1)

    def test_gaps(self):
        gaps1_all = self.series1.gaps(mode="all")
        self.assertTrue(gaps1_all.empty)
        gaps1_any = self.series1.gaps(mode="any")
        self.assertTrue(gaps1_any.empty)

        gaps4_all = self.series4.gaps(mode="all")
        self.assertTrue(
            (
                gaps4_all["gap_start"] == pd.DatetimeIndex([pd.Timestamp("20130208")])
            ).all()
        )
        self.assertTrue(
            (gaps4_all["gap_end"] == pd.DatetimeIndex([pd.Timestamp("20130208")])).all()
        )
        self.assertEqual(gaps4_all["gap_size"].values.tolist(), [1])

        gaps4_any = self.series4.gaps(mode="any")
        self.assertTrue(
            (
                gaps4_any["gap_start"]
                == pd.DatetimeIndex(
                    [pd.Timestamp("20130208"), pd.Timestamp("20130211")]
                )
            ).all()
        )
        self.assertTrue(
            (
                gaps4_any["gap_end"]
                == pd.DatetimeIndex(
                    [pd.Timestamp("20130208"), pd.Timestamp("20130214")]
                )
            ).all()
        )
        self.assertEqual(gaps4_any["gap_size"].values.tolist(), [1, 4])



"""
WassersteinScorer
-----

Wasserstein Scorer (distance function defined between probability distributions) [1]_.
The implementations is wrapped around `scipy.stats
<https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.wasserstein_distance.html>`_.

References
----------
.. [1] https://en.wikipedia.org/wiki/Wasserstein_metric
"""

from typing import Sequence

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from scipy.stats import wasserstein_distance

from darts.ad.scorers.scorers import FittableAnomalyScorer
from darts.logging import get_logger, raise_if_not
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class WassersteinScorer(FittableAnomalyScorer):
    def __init__(
        self,
        window: int = 10,
        component_wise: bool = False,
        diff_fn="abs_diff",
    ) -> None:
        """
        When calling ``fit(series)``, a moving window is applied, which results in a set of vectors of size `W`,
        where `W` is the window size. These vectors are kept in memory, representing the training
        distribution. The ``score(series)`` function will apply the same moving window.
        The Wasserstein distance is computed between the training distribution and each vector,
        resulting in an anomaly score.

        Alternatively, the scorer has the functions ``fit_from_prediction()`` and ``score_from_prediction()``.
        Both require two series (actual and prediction), and compute a "difference" series by applying the
        function ``diff_fn`` (default: absolute difference). The resulting series is then passed to the
        functions ``fit()`` and ``score()``, respectively.

        `component_wise` is a boolean parameter indicating how the model should behave with multivariate inputs
        series. If set to True, the model will treat each series dimension independently. If set to False, the model
        concatenates the dimensions in each windows of length `W` and computes a single score for all dimensions.

        **Training with** ``fit()``:

        The input can be a series (univariate or multivariate) or multiple series. The series will be partitioned
        into equal size subsequences. The subsequence will be of size `W` * `D`, with:

        * `W` being the size of the window given as a parameter `window`
        * `D` being the dimension of the series (`D` = 1 if univariate or if `component_wise` is set to True)

        For a series of length `N`, (`N` - `W` + 1)/W subsequences will be generated. If a list of series is given
        of length L, each series will be partitioned into subsequences, and the results will be concatenated into
        an array of length L * number of subsequences of each series.

        The arrays will be kept in memory, representing the training data distribution.
        In practice, the series or list of series can for instance represent residuals than can be
        considered independent and identically distributed (iid).

        If `component_wise` is set to True, the algorithm will be applied to each dimension independently. For each
        dimension, a PyOD model will be trained.

        **Computing score with** ``score()``:

        The input can be a series (univariate or multivariate) or a sequence of series. The given series must have the
        same dimension `D` as the data used to train the PyOD model.

        For each series, if the series is multivariate of dimension `D`:

        * if `component_wise` is set to False: it returns a univariate series (dimension=1). It represents
          the anomaly score of the entire series in the considered window at each timestamp.
        * if `component_wise` is set to True: it returns a multivariate series of dimension `D`. Each dimension
          represents the anomaly score of the corresponding component of the input.

        If the series is univariate, it returns a univariate series regardless of the parameter
        `component_wise`.

        A window of size `W` is rolled on the series with a stride equal to 1. It is the same size window `W` used
        during the training phase.
        Each value in the score series thus represents how anomalous the sample of the `W` previous values is.

        Parameters
        ----------
        window
            Size of the sliding window that represents the number of samples in the testing distribution to compare
            with the training distribution in the Wasserstein function
        diff_fn
            Optionally, reduced function to use if two series are given. It will transform the two series into one.
            This allows the WassersteinScorer to compute the Wasserstein distance on the original series or on its
            residuals (difference between the prediction and the original series).
            Must be one of "abs_diff" and "diff" (defined in ``_diff_series()``).
            Default: "abs_diff"
        component_wise
            Boolean value indicating if the score needs to be computed for each component independently (True)
            or by concatenating the component in the considered window to compute one score (False).
            Default: False

        """

        # TODO:
        #     - understand better the math behind the Wasserstein distance when the test distribution contains
        #     only one sample
        #     - check if there is an equivalent Wasserstein distance for d-D distributions (currently only accepts 1D)

        if type(window) is int:
            if window > 0 and window < 10:
                logger.warning(
                    f"The `window` parameter WassersteinScorer is smaller than 10 (w={window})."
                    + " The value represents the window length rolled on the series given as"
                    + " input in the ``score`` function. At each position, the w values will"
                    + " constitute a subset, and the Wasserstein distance between the subset"
                    + " and the train distribution will be computed. To better represent the"
                    + " constituted test distribution, the window parameter should be larger"
                    + " than 10."
                )

        raise_if_not(
            type(component_wise) is bool,
            f"Parameter `component_wise` must be Boolean, found type: {type(component_wise)}.",
        )
        self.component_wise = component_wise

        super().__init__(
            univariate_scorer=(not component_wise), window=window, diff_fn=diff_fn
        )

    def __str__(self):
        return "WassersteinScorer"

    def _fit_core(
        self,
        list_series: Sequence[TimeSeries],
    ):
        self.training_data = np.concatenate(
            [s.all_values(copy=False) for s in list_series]
        ).squeeze(-1)

        if not self.component_wise:
            self.training_data = self.training_data.flatten()

    def _score_core(self, series: TimeSeries) -> TimeSeries:
        raise_if_not(
            self.width_trained_on == series.width,
            "Input must have the same number of components as the data used for"
            + " training the Wasserstein model, found number of components equal"
            + f" to {series.width} and expected {self.width_trained_on}.",
        )

        np_series = series.all_values(copy=False)
        np_anomaly_score = []

        if not self.component_wise:
            np_anomaly_score = [
                wasserstein_distance(self.training_data, window_samples)
                for window_samples in sliding_window_view(
                    np_series, window_shape=self.window, axis=0
                )
                .transpose(0, 3, 1, 2)
                .reshape(-1, self.window * series.width)
            ]

            return TimeSeries.from_times_and_values(
                series.time_index[self.window - 1 :], np_anomaly_score
            )

        else:
            for component_idx in range(self.width_trained_on):
                score = [
                    wasserstein_distance(
                        self.training_data[component_idx, :], window_samples
                    )
                    for window_samples in sliding_window_view(
                        np_series[:, component_idx],
                        window_shape=self.window,
                        axis=0,
                    )
                    .transpose(0, 2, 1)
                    .reshape(-1, self.window)
                ]

                np_anomaly_score.append(score)

            return TimeSeries.from_times_and_values(
                series.time_index[self.window - 1 :], list(zip(*np_anomaly_score))
            )
